import 'dart:async';
import 'package:oobium_datastore/oobium_datastore.dart';
import 'package:oobium_logging/oobium_logging.dart';
part 'main.schema.g.feature_persist.dart';
part 'main.schema.x.dart';

final _log = Logger(MainData);

class MainData extends ModelContext {
  MainData({String? path, OnUpgrade? onUpgrade})
      : super(
            StateContext(features: [
              _ModelFeature(),
              if (path != null)
                _PersistFeature(
                    path: path,
                    version: 1,
                    compactionStrategy: const DefaultCompactionStrategy(),
                    converter: const _Converter(),
                    onUpgrade: onUpgrade ?? dropOnUpgrade)
            ]),
            const _Converter());

  Future<MainData> open() => super.open().then((_) => this);

  Items get items => feature<_ModelFeature>().items;

  void load(data) => feature<_ModelFeature>().load(data);
  void unload(data) => feature<_ModelFeature>().unload(data);
}

class _Converter extends Converter {
  const _Converter();

  @override
  StateValue decodePut(DataRecord r) {
    final data = r.toJsonData();
    switch (r.type) {
      case 'Item':
        return ItemSnap._(ItemKey.value(r.modelId), data['name'],
            ObjectId.fromHexString(r.updateId));
      default:
        throw Exception('unknown type ${r.type} during decodePut');
    }
  }

  @override
  StateKey decodeRemove(DataRecord r) {
    switch (r.type) {
      case 'Item':
        return ItemKey.parse(r.modelId);
      default:
        throw Exception('unknown type ${r.type} during decodeRemove');
    }
  }

  @override
  DataRecord encodePut(StateValue v) {
    if (v is ItemSnap) {
      return DataRecord(
          'Item', v.id.toString(), v.updateId.hexString, {'name': v.name});
    } else {
      throw Exception('unknown type $v during encodePut');
    }
  }

  @override
  DataRecord encodeRemove(StateValue v) {
    if (v is ItemSnap) {
      return DataRecord('Item', v.id.toString(), v.updateId.hexString);
    } else {
      throw Exception('unknown type $v during encodeRemove');
    }
  }
}

abstract class Item {
  int get id;
  String get name;
}

class ItemKey extends StateKey<ItemSnap> {
  @override
  final int id;
  const ItemKey(this.id);
  ItemKey.parse(String s) : id = ItemKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? ItemKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class ItemSnap extends Model<ItemSnap, int> implements Item, ItemKey {
  ItemSnap({required int id, required String name}) : this._(id, name);
  ItemSnap.fromJson(Map<String, dynamic> data,
      {int? id, String? name, ItemQuery? withQuery})
      : this._(
            id ?? (data['id'] as int), name ?? withQuery?.name ?? data['name']);
  static List<ItemSnap> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          String? name,
          ItemQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => ItemSnap.fromJson(data,
              id: id?.call(data), name: name, withQuery: withQuery))
          .toList();
  ItemSnap._(int id, this.name, [ObjectId? _updateId]) : super(id, _updateId);

  final String name;

  @override
  int compareTo(ItemSnap other) => compare(name, other.name) ?? 0;

  @override
  ItemSnap copyWith({int? id, String? name}) =>
      ItemSnap._(id ?? this.id, name ?? this.name);

  @override
  Map<String, dynamic> toJson() => {'id': id, 'name': name};
}

class ItemRef extends ModelRef<ItemSnap, int, ItemRef>
    implements Item, ItemKey {
  final _ModelFeature _feature;
  ItemRef._(this._feature, int id) : super(_feature, id);

  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);

  void update({String? name}) {
    final s = ItemSnap._(id, (name ?? this.name));
    _feature.put(s);
  }
}

class ItemQuery {
  final int? id;
  final String? name;

  const ItemQuery({this.id, this.name});
  ItemQuery._(this.id, this.name);

  bool get isAll => (id == null) && (name == null);
  bool get isNotAll => !isAll;

  bool test(ItemSnap m) =>
      (id == null || id == m.id) && (name == null || name == m.name);

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is ItemQuery && id == other.id && name == other.name);

  @override
  int get hashCode => Object.hash(Item, id, name);
}

class Items extends ModelCollection<Items, ItemSnap, int, ItemRef> {
  final _ModelFeature _feature;
  Items._(this._feature)
      : super(_feature, () => Items._(_feature),
            (s) => ItemRef._(_feature, s.id));
  ItemRef? operator [](int id) => getOrNull(id);
  ItemRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Item($id) does not exist'));
  ItemRef? getOrNull(int id) {
    final ref = ItemRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Iterable<ItemRef> find({int? id, String? name}) => _feature
      .getAll(ItemQuery._(id, name).test)
      .map((s) => ItemRef._(_feature, s.id));

  ItemRef put({required int id, required String name}) =>
      load(ItemSnap._(id, name));
}

class _ModelFeature extends WatchableStateFeature {
  late final items = Items._(this);
  void load(data) {
    if (data is ItemSnap) {
      put(data);
    } else if (data is Iterable) {
      data.forEach((e) => load(e));
    } else if (data is Map<String, dynamic>) {
      for (final key in data.keys) {
        if (key == 'item' && data[key] is Map<String, dynamic>) {
          load(ItemSnap.fromJson(data['item']));
        } else if (key == 'items' && data[key] is Iterable) {
          load(ItemSnap.fromJsonList(data['items']));
        } else {
          throw Exception(
              'invalid map entry during load: ($key, ${data[key]})');
        }
      }
    } else {
      throw Exception('attempted to load unknown type: $data');
    }
  }

  void unload(data) {
    if (data is ItemKey) {
      remove(data);
    } else if (data is Iterable) {
      data.forEach((e) => unload(e));
    } else if (data is Map<String, dynamic>) {
      for (final key in data.keys) {
        if (key == 'item' && data[key] is int) {
          unload(ItemKey(data[key] as int));
        } else if (key == 'items' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(ItemKey(data[key] as int)));
        } else {
          throw Exception(
              'invalid map entry during unload: ($key, ${data[key]})');
        }
      }
    } else {
      throw Exception('attempted to unload unknown type: $data');
    }
  }
}
