import 'package:objectid/objectid.dart';
import 'package:oobium_datastore/src/persist/feature.dart';
import 'package:oobium_datastore/src/state.dart';
import 'package:oobium_logging/oobium_logging.dart';
import 'package:test/test.dart';

import 'models_test.schema.g.dart';
import 'models_test.z.schema.g.dart' as z;

void main() async {
  Logger('oobium_datastore').child(StateStore).level = Level.trace;
  Logger(ModelsTestData).level = Level.trace;

  group('state', () {
    test('equality', () async {
      final ds = await ModelsTestData().open();
      final key = AccountKey();
      final model = AccountSnap(id: key.id, name: 'test-0');
      final ref = ds.accounts.load(model);
      print(key);
      print(model);
      print(ref);
      expect(key == model, isTrue);
      expect(model == ref, isTrue);
      expect(key == ref, isTrue);
    });

    test('hashcode', () async {
      final ds = await ModelsTestData().open();
      final key = AccountKey();
      final model = AccountSnap(id: key.id, name: 'test-0');
      final ref = ds.accounts.load(model);
      expect(key.hashCode, model.hashCode);
      expect(model.hashCode, ref.hashCode);
      expect(key.hashCode, ref.hashCode);
    });

    test('updateId', () async {
      final ds = await ModelsTestData().open();
      final key = AccountKey();

      final model = AccountSnap(id: key.id, name: 'test-0');
      expect(() => model.updateId, throwsException);

      final ref = ds.accounts.load(model);
      expect(model.updateId, isNotNull);
      expect(ref.updateId, isNotNull);
    });

    test('export', () async {
      final ds = await ModelsTestData().open();
      final i1 = ds.images.put(id: 1, height: 10, width: 100);
      final i2 = ds.images.put(id: 2, height: 20, width: 200);
      await pumpEventQueue();
      final lines = await ds.export().toList();
      expect(lines.length, 2);
      expect(lines[0], 'Image:1:${i1.updateId}{"height":10,"width":100}');
      expect(lines[1], 'Image:2:${i2.updateId}{"height":20,"width":200}');
    });

    test('import', () async {
      final ds = ModelsTestData();
      await ds.import(Stream.fromIterable([
        'Image:1:${ObjectId()}{"height":10,"width":100}',
        'Image:2:${ObjectId()}{"height":20,"width":200}',
      ]));
      expect(ds.images.length, 2);
    });

    test('import/export', () async {
      final ds = ModelsTestData();
      final line1 = 'Image:1:${ObjectId()}{"height":10,"width":100}';
      final line2 = 'Image:2:${ObjectId()}{"height":20,"width":200}';

      await ds.import(Stream.fromIterable([line1, line2,]));
      final lines = await ds.export().toList();

      expect(lines[0], line1);
      expect(lines[1], line2);
    });
  });

  group('load/unload', () {
    test('unknown objects throw a type error', () async {
      final ds = await ModelsTestData().open();
      expect(() => ds.load(['boom']), throwsException);
      expect(() => ds.unload(['boom']), throwsException);
    });

    test('snapshot', () async {
      final ds = await ModelsTestData().open();

      ds.load(ImageSnap(id: 123, name: 'image-1', height: 0, width: 0));
      expect(ds.images.length, 1);
      expect(ds.images[123], isNotNull);
      expect(ds.images[123]?.name, 'image-1');

      ds.unload(ImageSnap(id: 123, name: 'image-1', height: 0, width: 0));
      expect(ds.images.isEmpty, isTrue);
      expect(ds.images[123], isNull);
      expect(ds.images[123]?.name, isNull);
    });

    test('snapshots', () async {
      final ds = await ModelsTestData().open();

      ds.load([
        ImageSnap(id: 123, name: 'image-1', height: 0, width: 0),
        ImageSnap(id: 234, name: 'image-2', height: 0, width: 0),
      ]);
      expect(ds.images.length, 2);
      expect(ds.images[123], isNotNull);
      expect(ds.images[234], isNotNull);

      ds.unload([
        ImageSnap(id: 234, name: 'image-2', height: 0, width: 0),
      ]);
      expect(ds.images.length, 1);
      expect(ds.images[123], isNotNull);
      expect(ds.images[234], isNull);
    });

    test('load json', () async {
      final ds = await ModelsTestData().open();

      ds.load({
        'images': [
          {'id': 1, 'name': 'image-1', 'height': 2, 'width': 3},
          {'id': 4, 'name': 'image-2', 'height': 5, 'width': 6}
        ]
      });

      expect(ds.images.length, 2);
      expect(ds.images[1]?.name, 'image-1');
      expect(ds.images[4]?.name, 'image-2');
    });

    test('load related', () async {
      final ds = await ModelsTestData().open();

      ds.settings << SettingsSnap(
        userId: ds.users << UserSnap(id: 1, name: 'test-01', home: home)
      );

      expect(ds.settings.user?.name, 'test-01');
    });
  });

  group('singletons', () {
    test('singleton with default constructor', () async {
      final ds = await ModelsTestData().open();

      expect(ds.settings, isNotNull);
      expect(ds.settings.isInitialized, isTrue);
      expect(ds.settings.name, 'initial settings name');
      expect(ds.settings.user, isNull);
    });

    test('singleton without default constructor', () async {
      final ds = await ModelsTestData().open();

      expect(ds.config, isNotNull);
      expect(ds.config.isInitialized, isFalse);
      expect(() => ds.config.name, throwsException);
    });

    test('singleton without default constructor, preloaded', () async {
      final ds = await ModelsTestData().open();
      expect(ds.config.isInitialized, isFalse);

      ds.config.update(name: 'test-1', options: {});

      expect(ds.config.isInitialized, isTrue);
      expect(ds.config.name, 'test-1');
    });
  });

  group('put models', () {
    test('multiple with same external id', () async {
      final ds = await ModelsTestData().open();

      final image1 = ds.images.put(id: 1, height: 100, width: 200);
      expect(image1.width, 200);

      final image2 = ds.images.put(id: 1, height: 300, width: 400);
      expect(image1, image2);
      expect(image1, ds.images[1]);
      expect(image1.width, 400);
    });

    test('related', () async {
      final ds = await ModelsTestData().open();

      final user = ds.users.put(id: 1, name: 'user-1', home: home);
      final post = ds.posts.put(id: 2, title: 'test-1', userId: user.id);

      expect(ds.length, 2);

      expect(user.name, 'user-1');
      expect(user.posts.single, post);
      expect(ds.users.single, user);
      expect(ds.users[user.id], user);
      expect(ds.users.find(name: 'user-1').first, user);

      expect(post.title, 'test-1');
      expect(post.user, user);
      expect(ds.posts.first, post);
      expect(ds.posts[post.id], post);
      expect(ds.posts.find(title: 'test-1').first, post);
    });
  });

  group('get model', () {
    test('bracket accessor', () async {
      final ds = await ModelsTestData().open();
      ds.load([ImageSnap(id: 1, height: 10, width: 100)]);

      expect(ds.images[0], isNull);
      expect(ds.images[1], isNotNull);
    });

    test('get accessor', () async {
      final ds = await ModelsTestData().open();
      ds.load([ImageSnap(id: 1, height: 10, width: 100)]);

      expect(() => ds.images.get(0), throwsException);
      expect(ds.images.get(1), isNotNull);
    });

    test('getOrNull accessor', () async {
      final ds = await ModelsTestData().open();
      ds.load([ImageSnap(id: 1, height: 10, width: 100)]);

      expect(ds.images.getOrNull(0), isNull);
      expect(ds.images.getOrNull(1), isNotNull);
    });
  });

  group('update model', () {
    test('existing model, single field update', () async {
      final ds = await ModelsTestData().open();
      ds.load([ImageSnap(id: 1, height: 10, width: 100)]);

      ds.images[1]?.update(name: 'new name');

      expect(ds.images[1]?.name, 'new name');
    });

    test('existing model, single field setter', () async {
      final ds = await ModelsTestData().open();
      ds.load([ImageSnap(id: 1, height: 10, width: 100)]);

      ds.images[1]?.name = 'new name';

      expect(ds.images[1]?.name, 'new name');
    });

    test('existing model, single field update, null value', () async {
      final ds = await ModelsTestData().open();
      ds.load([ImageSnap(id: 1, name: 'test-01', height: 10, width: 100)]);

      expect(ds.images[1]?.name, 'test-01');
      ds.images[1]?.update(nameAsNull: true);
      expect(ds.images[1]?.name, isNull);
    });

    test('existing model, single field setter, null value', () async {
      final ds = await ModelsTestData().open();
      ds.load([ImageSnap(id: 1, name: 'test-01', height: 10, width: 100)]);

      expect(ds.images[1]?.name, 'test-01');
      ds.images[1]?.name = null;
      expect(ds.images[1]?.name, isNull);
      expect(ds.images[1]?.height, 10);
    });

    test('related models, loaded', () async {
      final ds = await ModelsTestData().open();
      ds.load([
        AccountSnap(name: 'test-account'),
        UserSnap(id: 1, name: 'test-user', home: home)
      ]);

      ds.settings.update(user: ds.users[1]);
      expect(ds.settings.user, ds.users[1]);

      ds.settings.update(account: ds.accounts.first);
      expect(ds.settings.account, ds.accounts.first);
    });
  });

  group('remove', () {
    test('remove parent with children', () async {
      final ds = await ModelsTestData().open();
      ds.load([
        UserSnap(id: 1, name: 'user-1', home: home),
        UserSnap(id: 2, name: 'user-2', home: home),
        UserSnap(id: 3, name: 'user-3', home: home),
        PostSnap(id: 1, title: 'post-1', userId: 1),
        PostSnap(id: 2, title: 'post-2', userId: 2),
        PostSnap(id: 3, title: 'post-3', userId: 3),
      ]);
      await pumpEventQueue();

      final removed = ds.users.get(2).remove();
      await pumpEventQueue();

      expect(removed, isNotNull);
      expect(removed?.name, 'user-2');
      expect(ds.users.length, 2);
      expect(ds.users.first.name, 'user-1');
      expect(ds.users.last.name, 'user-3');
      expect(ds.posts.length, 2);
      expect(ds.posts.first.title, 'post-1');
      expect(ds.posts.last.title, 'post-3');
    });

    test('remove collection of parents with children', () async {
      final ds = await ModelsTestData().open();
      ds.load([
        UserSnap(id: 1, name: 'user-1', home: home),
        UserSnap(id: 2, name: 'user-2', home: home),
        UserSnap(id: 3, name: 'user-3', home: home),
        PostSnap(id: 1, title: 'post-1', userId: 1),
        PostSnap(id: 2, title: 'post-2', userId: 2),
        PostSnap(id: 3, title: 'post-3', userId: 3),
      ]);
      await pumpEventQueue();

      final removed = ds.users.remove();
      await pumpEventQueue();

      expect(removed, isTrue);
      expect(ds.users, isEmpty);
      expect(ds.posts, isEmpty);
    });

    test('remove collection of parents with children, with where', () async {
      final ds = await ModelsTestData().open();
      ds.load([
        UserSnap(id: 1, name: 'user-1', home: home),
        UserSnap(id: 2, name: 'user-2', home: home),
        UserSnap(id: 3, name: 'user-3', home: home),
        PostSnap(id: 1, title: 'post-1', userId: 1),
        PostSnap(id: 2, title: 'post-2', userId: 2),
        PostSnap(id: 3, title: 'post-3', userId: 3),
      ]);
      await pumpEventQueue();

      final removed = ds.users.where((user) => user.id > 1).remove();
      await pumpEventQueue();

      expect(removed, isTrue);
      expect(ds.users.length, 1);
      expect(ds.users.single.name, 'user-1');
      expect(ds.posts.length, 1);
      expect(ds.posts.single.title, 'post-1');
    });

    test('remove collection of parents with children, with where, no results', () async {
      final ds = await ModelsTestData().open();
      ds.load([
        UserSnap(id: 1, name: 'user-1', home: home),
        UserSnap(id: 2, name: 'user-2', home: home),
        UserSnap(id: 3, name: 'user-3', home: home),
        PostSnap(id: 1, title: 'post-1', userId: 1),
        PostSnap(id: 2, title: 'post-2', userId: 2),
        PostSnap(id: 3, title: 'post-3', userId: 3),
      ]);
      await pumpEventQueue();

      final removed = ds.users.where((user) => user.id > 3).remove();
      await pumpEventQueue();

      expect(removed, isFalse);
      expect(ds.users.length, 3);
      expect(ds.posts.length, 3);
    });
  });

  group('set', () {
    test('set parent with children', () async {
      final ds = await ModelsTestData().open();
      ds.load([
        UserSnap(id: 1, name: 'user-1', home: home),
        UserSnap(id: 2, name: 'user-2', home: home),
        UserSnap(id: 3, name: 'user-3', home: home),
        PostSnap(id: 1, title: 'post-1', userId: 1),
        PostSnap(id: 2, title: 'post-2', userId: 2),
        PostSnap(id: 3, title: 'post-3', userId: 3),
      ]);
      await pumpEventQueue();

      expect(ds.length, 6);
      expect(ds.users.length, 3);
      expect(ds.posts.length, 3);

      final set = ds.users.set([UserSnap(id: 2, name: 'user-2', home: home)]);
      await pumpEventQueue();

      expect(ds.length, 2);
      expect(ds.users.length, 1);
      expect(ds.posts.length, 1);
      expect(ds.users.single.name, 'user-2');
      expect(ds.posts.single.title, 'post-2');
      expect(set.length, 1);
      expect(set.single.name, 'user-2');
      expect(set.single.posts.length, 1);
      expect(set.single.posts.single.title, 'post-2');
    });

    test('set parent with children and grandchildren', () async {
      final ds = await ModelsTestData().open();
      ds.load([
        UserSnap(id: 1, name: 'user-1', home: home),
        UserSnap(id: 2, name: 'user-2', home: home),
        UserSnap(id: 3, name: 'user-3', home: home),
        PostSnap(id: 1, title: 'post-1', userId: 1),
        PostSnap(id: 2, title: 'post-2', userId: 2),
        PostSnap(id: 3, title: 'post-3', userId: 3),
        CommentSnap(id: 1, message: 'comment-1', postId: 1, ownerId: 1),
        CommentSnap(id: 2, message: 'comment-2', postId: 2, ownerId: 2),
        CommentSnap(id: 3, message: 'comment-3', postId: 3, ownerId: 3),
      ]);
      await pumpEventQueue();

      expect(ds.length, 9);
      expect(ds.users.length, 3);
      expect(ds.posts.length, 3);
      expect(ds.comments.length, 3);

      final set = ds.users.set([UserSnap(id: 2, name: 'user-2', home: home)]);
      await pumpEventQueue();

      expect(ds.length, 3);
      expect(ds.users.length, 1);
      expect(ds.posts.length, 1);
      expect(ds.comments.length, 1);
      expect(ds.users.single.name, 'user-2');
      expect(ds.posts.single.title, 'post-2');
      expect(ds.comments.single.message, 'comment-2');
      expect(set.length, 1);
      expect(set.single.name, 'user-2');
      expect(set.single.posts.length, 1);
      expect(set.single.posts.single.title, 'post-2');
      expect(set.single.posts.single.comments.length, 1);
      expect(set.single.posts.single.comments.single.message, 'comment-2');
    });
  });

  group('operations', () {
    test('puts', () async {
      final ds = await ModelsTestData().open();

      // final s = ds.watch().where((v) => v is User && v.id == 1).map((u) => ds.users.get(1));
      // ds.users.where((u) => u.id == 1).watch();
      // ds.posts.where((p) => p.user == ds.users[1]).watch();
      ds.users.watch().listen(expectAsync1((e) {
        expect(e.length, 2);
      }, count: 1));

      ds.users.put(id: 1, name: 'user-1', home: home);
      ds.users.put(id: 2, name: 'user-2', home: home);
      ds.posts.put(id: 3, title: 'post-3', userId: 1);
    });

    test('updates', () async {
      final ds = await ModelsTestData().open();

      ds.users.watch().listen(expectAsync1((e) {
        expect(e.length, 1);
      }, count: 1));

      final user = ds.users.put(id: 1, name: 'user-1', home: home);
      user.name = 'user-2';
      user.name = 'user-3';
    });

    test('updates, with notify', () async {
      final ds = await ModelsTestData().open();

      ds.users.watch().listen(expectAsync1((e) {
        expect(e.length, 1);
      }, count: 2));

      final user = ds.users.put(id: 1, name: 'user-1', home: home);
      user.name = 'user-2';
      user.notify();
      user.name = 'user-3';
    });

    test('updates, with flush', () async {
      final ds = await ModelsTestData().open();

      ds.users.watch().listen(expectAsync1((e) {
        expect(e.length, 1);
      }, count: 2));

      final user = ds.users.put(id: 1, name: 'user-1', home: home);
      user.name = 'user-2';
      await ds.flush();
      user.name = 'user-3';
    });
  });

  group('fetch', () {
    test('simple without id (fetch all)', () async {
      final ds = await ModelsTestData(
        imageRestAdapter: ImageRestAdapter(
          onRequest: expectAsync2((ds, query) {
            expect(query.isAll, isTrue);
          }, count: 1),
        )).open();

      await ds.images.fetch();
      await ds.images.fetch();
    });

    test('expire simple without id (fetch all)', () async {
      final ds = await ModelsTestData(
        imageRestAdapter: ImageRestAdapter(
          onRequest: expectAsync2((ds, query) {
            expect(query.isAll, isTrue);
          }, count: 2),
        )).open();

      await ds.images.fetch();
      ds.images.expire();
      await ds.images.fetch();
    });

    test('simple with id (fetch one)', () async {
      final ds = await ModelsTestData(
        imageRestAdapter: ImageRestAdapter(
          onRequest: expectAsync2((ds, query) {
            expect(query.isAll, isFalse);
            expect(query.id, 123);
            return {'id': 123, 'name': 'image-1', 'width': 1, 'height': 2};
          }, count: 1),
        )).open();

      await ds.images.fetch(id: 123);
      await ds.images.fetch(id: 123);

      expect(ds.images[321], isNull);
      expect(ds.images[123], isNotNull);
      expect(ds.images[123]?.name, 'image-1');
    });

    test('expire simple with id (fetch one)', () async {
      final ds = await ModelsTestData(
        imageRestAdapter: ImageRestAdapter(
          onRequest: expectAsync2((ds, query) {
            expect(query.isAll, isFalse);
            expect(query.id, 123);
            return {'id': 123, 'name': 'image-1', 'width': 1, 'height': 2};
          }, count: 2),
        )).open();

      await ds.images.fetch(id: 123);
      ds.images.expire(id: 123);
      await ds.images.fetch(id: 123);
    });

    test('singleton', () async {
      final ds = await ModelsTestData(
        configRestAdapter: ConfigRestAdapter(
          onRequest: expectAsync1((ds) =>
            {'id': 'whatever', 'name': 'test-2'}, count: 1
          ),
        )).open();

      await ds.config.fetch();

      expect(ds.config.id, 'Config');
      expect(ds.config, isNotNull);
      expect(ds.config.name, 'test-2');
    });

    test('related with id (fetch one)', () async {
      final ds = await ModelsTestData(
        userRestAdapter: UserRestAdapter(
          onRequest: expectAsync2((ds, query) {
            expect(query.isAll, isFalse);
            expect(query.id, 123);
            return {'id': 123, 'name': 'test-2', 'home': home.toJson()};
          }, count: 1),
        ),
        postRestAdapter: PostRestAdapter(
          onRequest: expectAsync2((ds, query) {
            expect(query.isAll, isFalse);
            expect(query.userId, 123);
            return [
              {'id': 234, 'title': 'title-1', 'user': 123},
              {'id': 345, 'title': 'title-2', 'user': 123},
            ];
          }, count: 1)
        ),
        commentRestAdapter: CommentRestAdapter(
          onRequest: expectAsync2((ds, query) {
            expect(query.isAll, isFalse);
            expect(query.postId == 234 || query.postId == 345, isTrue);
            return [];
          }, count: 2)
        )
      ).open();

      await ds.users.fetch(id: 123);
      await ds.users[123]?.posts.fetch();
      await ds.users[123]?.posts[234]?.comments.fetch();
      await ds.users[123]?.posts[345]?.comments.fetch();

      expect(ds.users[123], isNotNull);
      expect(ds.users[123]?.id, 123);
      expect(ds.users[123]?.name, 'test-2');
      expect(ds.posts.length, 2);
      expect(ds.users[123]?.posts.length, 2);
      expect(ds.posts[234]?.title, 'title-1');
      expect(ds.posts[345]?.title, 'title-2');
    });

    test('parent without id (fetch all)', () async {
      final ds = await ModelsTestData(
          userRestAdapter: UserRestAdapter(
            onRequest: expectAsync2((ds, query) {
              expect(query.isAll, isTrue);
              return [
                {'id': 123, 'name': 'user-1', 'home': home.toJson()},
                {'id': 234, 'name': 'user-2', 'home': home.toJson()},
              ];
            }, count: 1),
          ),
          postRestAdapter: PostRestAdapter(
            onRequest: expectAsync2((ds, query) {
              expect(query.isAll, isFalse);
              return [];
            }, count: 2)
          )
      )
      .open();

      await ds.users.fetch();
      await ds.users[123]?.posts.fetch();
      await ds.users[234]?.posts.fetch();

      expect(ds.users.length, 2);
      expect(ds.users[123]?.name, 'user-1');
      expect(ds.users[234]?.name, 'user-2');
    });

    test('child without id (fetch all)', () async {
      final ds = await ModelsTestData(
        postRestAdapter: PostRestAdapter(
          onRequest: expectAsync2((ds, query) {
            expect(query.isAll, isFalse);
            expect(query.userId, 1);
            return [
              {'id': 123, 'title': 'post-1', 'user': 1},
              {'id': 234, 'title': 'post-2', 'user': 1},
            ];
          }, count: 1),
        ),
        commentRestAdapter: CommentRestAdapter(
          onRequest: expectAsync2((ds, query) {
            expect(query.isAll, isFalse);
            return [];
          }, count: 2)
        )
      ).open();
      final user = ds.users.put(id: 1, name: 'user-1', home: home);

      await ds.posts.fetch(userId: user.id);
      await ds.posts[123]?.comments.fetch();
      await ds.posts[234]?.comments.fetch();

      expect(ds.posts.length, 2);
      expect(ds.posts[123]?.title, 'post-1');
      expect(ds.posts[234]?.title, 'post-2');
    });
  });

  group('rest adapters', () {
    test('onCreated', () async {
      final ds = await ModelsTestData(
        userRestAdapter: UserRestAdapter(
          onCreated: expectAsync2((ds, user) {
            expect(user.id, 1);
            expect(user.name, 'user-1');
          })
        )
      ).open();
      ds.users.put(id: 1, name: 'user-1', home: home).id;
    });

    test('onUpdated', () async {
      final ds = await ModelsTestData(
        userRestAdapter: UserRestAdapter(
          onUpdated: expectAsync3((ds, oldUser, user) {
            expect(oldUser.id, 1);
            expect(oldUser.name, 'user-1');
            expect(user.id, 1);
            expect(user.name, 'user-2');
          })
        )
      ).open();
      ds.users << UserSnap(id: 1, name: 'user-1', home: home);
      await pumpEventQueue();
      ds.users.put(id: 1, name: 'user-2', home: home);
    });

    test('onRemoved', () async {
      final ds = await ModelsTestData(
        userRestAdapter: UserRestAdapter(
          onRemoved: expectAsync2((ds, user) {
            expect(user.id, 1);
            expect(user.name, 'user-1');
          })
        )
      ).open();
      ds.users << UserSnap(id: 1, name: 'user-1', home: home);
      await pumpEventQueue();
      ds.users[1]?.remove();
    });
  });

  group('persist', () {
    test('<<', () async {
      final ds = await ModelsTestData(path: _path('test-data')).open();
      ds.users << UserSnap(id: 1, name: 'test-user-1', home: home);
      await ds.close();
      await ds.open();
      expect(ds.users.length, 1);
    });

    test('load', () async {
      final ds = await ModelsTestData(path: _path('test-data')).open();
      ds.users.load(UserSnap(id: 1, name: 'test-user-1', home: home));
      await ds.close();
      await ds.open();
      expect(ds.users.length, 1);
      expect(ds.users.first.name, 'test-user-1');
    });

    test('put', () async {
      final ds = await ModelsTestData(path: _path('test-data')).open();
      ds.users.put(id: 1, name: 'test-user-1', home: home);
      await ds.close();
      await ds.open();
      expect(ds.users.length, 1);
      expect(ds.users.first.name, 'test-user-1');
    });

    test('put update', () async {
      final ds = await ModelsTestData(path: _path('test-data')).open();

      final user = ds.users.put(id: 1, name: 'test-user-1', home: home);
      user.name = 'test-user-2';

      await ds.close();
      await ds.open();

      expect(ds.users.length, 1);
      expect(ds.users.first.name, 'test-user-2');
    });

    test('put remove, immediate', () async {
      final ds = await ModelsTestData(path: _path('test-data')).open();

      final user = ds.users.put(id: 1, name: 'test-user-1', home: home);
      user.remove();

      await ds.close();
      await ds.open();

      expect(ds.users, isEmpty);
    });

    test('put remove, with delay', () async {
      final ds = await ModelsTestData(path: _path('test-data')).open();

      final user = ds.users.put(id: 1, name: 'test-user-1', home: home);
      await pumpEventQueue();
      user.remove();

      await ds.close();
      await ds.open();

      expect(ds.users, isEmpty);
    });

    test('put remove, with load', () async {
      final ds = await ModelsTestData(path: _path('test-data')).open();

      ds.users.put(id: 1, name: 'test-user-1', home: home);

      await ds.close();
      await ds.open();

      ds.users.get(1).remove();

      expect(ds.users, isEmpty);
    });
  });

  group('upgrade', () {
    test('drop (default)', () async {
      final path = _path('test-data');
      final ds = await ModelsTestData(path: path).open();
      ds.images.put(id: 1, height: 10, width: 100);
      await ds.close();

      final zs = await z.ModelsTestData(path: path).open();
      expect(zs, isEmpty);
    });

    test('pass through', () async {
      final path = _path('test-data');
      final ds = await ModelsTestData(path: path).open();
      ds.images.put(id: 1, height: 10, width: 100);
      await ds.close();

      final zs = await z.ModelsTestData(
          path: path,
          onUpgrade: (e) => e.oldData
      );

      expect(() => zs.open(), throwsA(isA<TypeError>()));
    });

    test('convert', () async {
      final path = _path('test-data');
      final ds = await ModelsTestData(path: path).open();
      ds.images.put(id: 1, height: 10, width: 100);
      await ds.close();

      final zs = await z.ModelsTestData(
          path: path,
          onUpgrade: (e) => e.oldData.map((record) =>
              record.withData(record.toJsonData()..['name'] = 'upgrade')
          )
      ).open();

      expect(zs.images.length, 1);
      expect(zs.images.first.name, 'upgrade');
    });

    test('version z', () async {
      final path = _path('test-data');
      final ds = await ModelsTestData(path: path).open();
      ds.images.put(id: 1, height: 10, width: 100);
      await ds.close();

      final zs = await z.ModelsTestData(
        path: path,
        onUpgrade: expectAsync1((event) {
          expect(event.oldVersion, 1);
          expect(event.newVersion, 3);
          return Stream<DataRecord>.empty();
        }, count: 1)
      ).open();

      expect(zs, isEmpty);
    });
  });
}

var count = 0;
String _path(String root) => '$root/test-${count++}';

Address get home => Address(
    street1: 'Main St',
    city: 'AnyCity',
    state: 'AnyState',
    zipcode: '01234'
);

