import 'dart:async';
import 'package:oobium_datastore/oobium_datastore.dart';
import 'package:oobium_logging/oobium_logging.dart';
part 'models_test.schema.g.feature_persist.dart';
part 'models_test.schema.g.feature_rest.dart';
part 'models_test.schema.g.feature_trigger.dart';
part 'models_test.schema.x.dart';

final _log = Logger(ModelsTestData);

class ModelsTestData extends ModelContext {
  ModelsTestData(
      {String? path,
      OnUpgrade? onUpgrade,
      SettingsRestAdapter? settingsRestAdapter,
      ConfigRestAdapter? configRestAdapter,
      AccountRestAdapter? accountRestAdapter,
      ImageRestAdapter? imageRestAdapter,
      UserRestAdapter? userRestAdapter,
      PostRestAdapter? postRestAdapter,
      CommentRestAdapter? commentRestAdapter,
      SettingsTriggers? settingsTriggers,
      ConfigTriggers? configTriggers,
      AccountTriggers? accountTriggers,
      ImageTriggers? imageTriggers,
      UserTriggers? userTriggers,
      PostTriggers? postTriggers,
      CommentTriggers? commentTriggers})
      : super(
            StateContext(features: [
              _RestFeature(
                settingsRestAdapter:
                    settingsRestAdapter ?? SettingsRestAdapter(),
                configRestAdapter: configRestAdapter ?? ConfigRestAdapter(),
                accountRestAdapter: accountRestAdapter ?? AccountRestAdapter(),
                imageRestAdapter: imageRestAdapter ?? ImageRestAdapter(),
                userRestAdapter: userRestAdapter ?? UserRestAdapter(),
                postRestAdapter: postRestAdapter ?? PostRestAdapter(),
                commentRestAdapter: commentRestAdapter ?? CommentRestAdapter(),
              ),
              _TriggerFeature(
                settingsTriggers: settingsTriggers,
                configTriggers: configTriggers,
                accountTriggers: accountTriggers,
                imageTriggers: imageTriggers,
                userTriggers: userTriggers,
                postTriggers: postTriggers,
                commentTriggers: commentTriggers,
              ),
              _ModelFeature(),
              if (path != null)
                _PersistFeature(
                    path: path,
                    version: 1,
                    compactionStrategy: const DefaultCompactionStrategy(),
                    converter: const _Converter(),
                    onUpgrade: onUpgrade ?? dropOnUpgrade)
            ]),
            const _Converter()) {
    feature<_RestFeature>()._modelContext = this;
  }

  Future<ModelsTestData> open() => super.open().then((_) => this);

  SettingsRef get settings => feature<_ModelFeature>().settings;
  ConfigRef get config => feature<_ModelFeature>().config;
  Accounts get accounts => feature<_ModelFeature>().accounts;
  Images get images => feature<_ModelFeature>().images;
  Users get users => feature<_ModelFeature>().users;
  Posts get posts => feature<_ModelFeature>().posts;
  Comments get comments => feature<_ModelFeature>().comments;

  void load(data) => feature<_ModelFeature>().load(data);
  void unload(data) => feature<_ModelFeature>().unload(data);
}

class _Converter extends Converter {
  const _Converter();

  @override
  StateValue decodePut(DataRecord r) {
    final data = r.toJsonData();
    switch (r.type) {
      case 'Settings':
        return SettingsSnap._(
            data['name'],
            ((data['account'] != null)
                ? AccountKey.value(data['account'])
                : null),
            ((data['user'] != null) ? data['user'] : null),
            ObjectId.fromHexString(r.updateId));
      case 'Config':
        return ConfigSnap._(
            data['name'],
            (data['options'] ?? {})
                .map((k, v) => MapEntry('$k', v))
                .cast<String, dynamic>(),
            ObjectId.fromHexString(r.updateId));
      case 'Account':
        return AccountSnap._(
            AccountKey.value(r.modelId),
            data['name'],
            ((data['settings'] != null) ? data['settings'] : null),
            ObjectId.fromHexString(r.updateId));
      case 'Image':
        return ImageSnap._(
            ImageKey.value(r.modelId),
            ((data['name'] != null) ? data['name'] : null),
            data['height'],
            data['width'],
            ObjectId.fromHexString(r.updateId));
      case 'User':
        return UserSnap._(
            UserKey.value(r.modelId),
            data['name'],
            ((data['desc'] != null) ? data['desc'] : null),
            Address.fromJson(data['home']),
            ((data['work'] != null) ? Address.fromJson(data['work']) : null),
            ObjectId.fromHexString(r.updateId));
      case 'Post':
        return PostSnap._(PostKey.value(r.modelId), data['title'], data['user'],
            ObjectId.fromHexString(r.updateId));
      case 'Comment':
        return CommentSnap._(
            CommentKey.value(r.modelId),
            data['post'],
            data['message'],
            data['owner'],
            ((data['auditor'] != null) ? data['auditor'] : null),
            ObjectId.fromHexString(r.updateId));
      default:
        throw Exception('unknown type ${r.type} during decodePut');
    }
  }

  @override
  StateKey decodeRemove(DataRecord r) {
    switch (r.type) {
      case 'Settings':
        return SettingsKey();
      case 'Config':
        return ConfigKey();
      case 'Account':
        return AccountKey.parse(r.modelId);
      case 'Image':
        return ImageKey.parse(r.modelId);
      case 'User':
        return UserKey.parse(r.modelId);
      case 'Post':
        return PostKey.parse(r.modelId);
      case 'Comment':
        return CommentKey.parse(r.modelId);
      default:
        throw Exception('unknown type ${r.type} during decodeRemove');
    }
  }

  @override
  DataRecord encodePut(StateValue v) {
    if (v is SettingsSnap) {
      return DataRecord('Settings', 'Settings', v.updateId.hexString, {
        'name': v.name,
        if (v.accountId != null) 'account': v.accountId?.hexString,
        if (v.userId != null) 'user': v.userId
      });
    } else if (v is ConfigSnap) {
      return DataRecord('Config', 'Config', v.updateId.hexString,
          {'name': v.name, 'options': recordValue(v.options)});
    } else if (v is AccountSnap) {
      return DataRecord('Account', v.id.toString(), v.updateId.hexString,
          {'name': v.name, if (v.settingsId != null) 'settings': v.settingsId});
    } else if (v is ImageSnap) {
      return DataRecord('Image', v.id.toString(), v.updateId.hexString, {
        if (v.name != null) 'name': v.name,
        'height': v.height,
        'width': v.width
      });
    } else if (v is UserSnap) {
      return DataRecord('User', v.id.toString(), v.updateId.hexString, {
        'name': v.name,
        if (v.desc != null) 'desc': v.desc,
        'home': v.home.toJson(),
        if (v.work != null) 'work': v.work?.toJson()
      });
    } else if (v is PostSnap) {
      return DataRecord('Post', v.id.toString(), v.updateId.hexString,
          {'title': v.title, 'user': v.userId});
    } else if (v is CommentSnap) {
      return DataRecord('Comment', v.id.toString(), v.updateId.hexString, {
        'post': v.postId,
        'message': v.message,
        'owner': v.ownerId,
        if (v.auditorId != null) 'auditor': v.auditorId
      });
    } else {
      throw Exception('unknown type $v during encodePut');
    }
  }

  @override
  DataRecord encodeRemove(StateValue v) {
    if (v is SettingsSnap) {
      return DataRecord('Settings', 'Settings', v.updateId.hexString);
    } else if (v is ConfigSnap) {
      return DataRecord('Config', 'Config', v.updateId.hexString);
    } else if (v is AccountSnap) {
      return DataRecord('Account', v.id.toString(), v.updateId.hexString);
    } else if (v is ImageSnap) {
      return DataRecord('Image', v.id.toString(), v.updateId.hexString);
    } else if (v is UserSnap) {
      return DataRecord('User', v.id.toString(), v.updateId.hexString);
    } else if (v is PostSnap) {
      return DataRecord('Post', v.id.toString(), v.updateId.hexString);
    } else if (v is CommentSnap) {
      return DataRecord('Comment', v.id.toString(), v.updateId.hexString);
    } else {
      throw Exception('unknown type $v during encodeRemove');
    }
  }
}

class Address implements Comparable<Address> {
  Address(
      {required String street1,
      String? street2,
      required String city,
      required String state,
      required String zipcode,
      Map<String, dynamic> extras = const {}})
      : this._(street1, street2, city, state, zipcode, extras);
  Address.fromJson(Map<String, dynamic> data,
      {String? street1,
      String? street2,
      bool street2AsNull = false,
      String? city,
      String? state,
      String? zipcode,
      Map<String, dynamic>? extras})
      : this._(
            street1 ?? data['street1'],
            street2 ?? ((data['street2'] != null) ? data['street2'] : null),
            city ?? data['city'],
            state ?? data['state'],
            zipcode ?? data['zipcode'],
            extras ??
                ((data['extras'] != null)
                    ? (data['extras'] ?? {})
                        .map((k, v) => MapEntry('$k', v))
                        .cast<String, dynamic>()
                    : const {}));
  static List<Address> fromJsonList(list,
          {String? street1,
          String? street2,
          bool street2AsNull = false,
          String? city,
          String? state,
          String? zipcode,
          Map<String, dynamic>? extras}) =>
      (list as Iterable)
          .map((data) => Address.fromJson(data,
              street1: street1,
              street2: street2,
              street2AsNull: street2AsNull,
              city: city,
              state: state,
              zipcode: zipcode,
              extras: extras))
          .toList();
  Address._(this.street1, this.street2, this.city, this.state, this.zipcode,
      this.extras);

  final String street1;
  final String? street2;
  final String city;
  final String state;
  final String zipcode;
  final Map<String, dynamic> extras;

  @override
  int compareTo(Address other) =>
      compare(street1, other.street1) ??
      compare(street2, other.street2) ??
      compare(city, other.city) ??
      compare(state, other.state) ??
      compare(zipcode, other.zipcode) ??
      compare(extras, other.extras) ??
      0;

  Address copyWith(
          {String? street1,
          String? street2,
          bool street2AsNull = false,
          String? city,
          String? state,
          String? zipcode,
          Map<String, dynamic>? extras}) =>
      Address._(
          street1 ?? this.street1,
          street2AsNull ? null : (street2 ?? this.street2),
          city ?? this.city,
          state ?? this.state,
          zipcode ?? this.zipcode,
          extras ?? this.extras);

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is Address &&
          (street1 == other.street1 ||
              street2 == other.street2 ||
              city == other.city ||
              state == other.state ||
              zipcode == other.zipcode ||
              extras == other.extras));

  @override
  int get hashCode =>
      Object.hash(Address, street1, street2, city, state, zipcode, extras);

  Map<String, dynamic> toJson() => {
        'street1': street1,
        if (street2 != null) 'street2': street2,
        'city': city,
        'state': state,
        'zipcode': zipcode,
        'extras': jsonValue(extras)
      };
}

abstract class Settings {
  String get name;
  ObjectId? get accountId;
  int? get userId;
}

class SettingsKey extends StateKey<SettingsSnap> {
  @override
  final String id = 'Settings';
  const SettingsKey();
}

class SettingsSnap extends Singleton<SettingsSnap>
    implements Settings, SettingsKey {
  SettingsSnap(
      {String name = 'initial settings name', ObjectId? accountId, int? userId})
      : this._(name, accountId, userId);
  SettingsSnap.fromJson(Map<String, dynamic> data,
      {String? name,
      ObjectId? accountId,
      bool accountIdAsNull = false,
      int? userId,
      bool userIdAsNull = false})
      : this._(
            name ??
                ((data['name'] != null)
                    ? data['name']
                    : 'initial settings name'),
            accountId ??
                ((data['account'] != null)
                    ? AccountKey.value(data['account'])
                    : null),
            userId ?? ((data['user'] != null) ? data['user'] : null));
  static List<SettingsSnap> fromJsonList(list,
          {String? name,
          ObjectId? accountId,
          bool accountIdAsNull = false,
          int? userId,
          bool userIdAsNull = false}) =>
      (list as Iterable)
          .map((data) => SettingsSnap.fromJson(data,
              name: name,
              accountId: accountId,
              accountIdAsNull: accountIdAsNull,
              userId: userId,
              userIdAsNull: userIdAsNull))
          .toList();
  SettingsSnap._(this.name, this.accountId, this.userId, [ObjectId? _updateId])
      : super(_updateId);

  final String id = 'Settings';
  final String name;
  final ObjectId? accountId;
  final int? userId;

  @override
  int compareTo(SettingsSnap other) =>
      compare(name, other.name) ??
      compare(accountId, other.accountId) ??
      compare(userId, other.userId) ??
      0;

  SettingsSnap copyWith(
          {String? name,
          ObjectId? accountId,
          bool accountIdAsNull = false,
          int? userId,
          bool userIdAsNull = false}) =>
      SettingsSnap._(
          name ?? this.name,
          accountIdAsNull ? null : (accountId ?? this.accountId),
          userIdAsNull ? null : (userId ?? this.userId));

  @override
  Map<String, dynamic> toJson() => {
        'id': id,
        'name': name,
        if (accountId != null) 'account': accountId?.hexString,
        if (userId != null) 'user': userId
      };
}

class SettingsRef extends SingletonRef<SettingsSnap, SettingsRef>
    implements Settings, SettingsKey {
  final _ModelFeature _feature;
  SettingsRef._(this._feature) : super(_feature);

  final String id = 'Settings';
  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);
  AccountRef? get account {
    final id = _feature.get(this).accountId;
    return (id != null) ? AccountRef._(_feature, id) : null;
  }

  ObjectId? get accountId => _feature.get(this).accountId;
  set account(AccountRef? value) =>
      (value != null) ? update(account: value) : update(accountAsNull: true);
  UserRef? get user {
    final id = _feature.get(this).userId;
    return (id != null) ? UserRef._(_feature, id) : null;
  }

  int? get userId => _feature.get(this).userId;
  set user(UserRef? value) =>
      (value != null) ? update(user: value) : update(userAsNull: true);

  void expire() => _feature<_RestFeature>().settingsRestAdapter.expire();

  Future<SettingsRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .settingsRestAdapter
      .fetch(force)
      .then((_) => this);

  void update(
      {String? name,
      AccountKey? account,
      bool accountAsNull = false,
      UserKey? user,
      bool userAsNull = false}) {
    final s = SettingsSnap._(
        (name ?? this.name),
        accountAsNull ? null : (account ?? this.account)?.id,
        userAsNull ? null : (user ?? this.user)?.id);
    _feature.put(s);
  }
}

abstract class Config {
  String get name;
  Map<String, dynamic> get options;
}

class ConfigKey extends StateKey<ConfigSnap> {
  @override
  final String id = 'Config';
  const ConfigKey();
}

class ConfigSnap extends Singleton<ConfigSnap> implements Config, ConfigKey {
  ConfigSnap({required String name, Map<String, dynamic> options = const {}})
      : this._(name, options);
  ConfigSnap.fromJson(Map<String, dynamic> data,
      {String? name, Map<String, dynamic>? options})
      : this._(
            name ?? data['name'],
            options ??
                ((data['options'] != null)
                    ? (data['options'] ?? {})
                        .map((k, v) => MapEntry('$k', v))
                        .cast<String, dynamic>()
                    : const {}));
  static List<ConfigSnap> fromJsonList(list,
          {String? name, Map<String, dynamic>? options}) =>
      (list as Iterable)
          .map(
              (data) => ConfigSnap.fromJson(data, name: name, options: options))
          .toList();
  ConfigSnap._(this.name, this.options, [ObjectId? _updateId])
      : super(_updateId);

  final String id = 'Config';
  final String name;
  final Map<String, dynamic> options;

  @override
  int compareTo(ConfigSnap other) =>
      compare(name, other.name) ?? compare(options, other.options) ?? 0;

  ConfigSnap copyWith({String? name, Map<String, dynamic>? options}) =>
      ConfigSnap._(name ?? this.name, options ?? this.options);

  @override
  Map<String, dynamic> toJson() =>
      {'id': id, 'name': name, 'options': jsonValue(options)};
}

class ConfigRef extends SingletonRef<ConfigSnap, ConfigRef>
    implements Config, ConfigKey {
  final _ModelFeature _feature;
  ConfigRef._(this._feature) : super(_feature);

  final String id = 'Config';
  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);
  Map<String, dynamic> get options => _feature.get(this).options;
  set options(Map<String, dynamic> value) => update(options: value);

  void expire() => _feature<_RestFeature>().configRestAdapter.expire();

  Future<ConfigRef> fetch({bool force = false}) =>
      _feature<_RestFeature>().configRestAdapter.fetch(force).then((_) => this);

  void update({String? name, Map<String, dynamic>? options}) {
    final s = ConfigSnap._((name ?? this.name), (options ?? this.options));
    _feature.put(s);
  }
}

abstract class Account {
  ObjectId get id;
  String get name;
  String? get settingsId;
}

class AccountKey extends StateKey<AccountSnap> {
  @override
  final ObjectId id;
  AccountKey([ObjectId? id]) : id = id ?? ObjectId();
  AccountKey.parse(String s) : id = AccountKey.value(s);
  static ObjectId value(String s) => ObjectId.fromHexString(s);
  static ObjectId? maybe(String? s) {
    try {
      return (s != null) ? AccountKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class AccountSnap extends Model<AccountSnap, ObjectId>
    implements Account, AccountKey {
  AccountSnap({ObjectId? id, required String name, String? settingsId})
      : this._(id ?? ObjectId(), name, settingsId);
  AccountSnap.fromJson(Map<String, dynamic> data,
      {ObjectId? id,
      String? name,
      String? settingsId,
      bool settingsIdAsNull = false,
      AccountQuery? withQuery})
      : this._(
            id ??
                ((data['id'] != null)
                    ? AccountKey.value(data['id'])
                    : ObjectId()),
            name ?? withQuery?.name ?? data['name'],
            settingsId ??
                withQuery?.settingsId ??
                ((data['settings'] != null) ? data['settings'] : null));
  static List<AccountSnap> fromJsonList(list,
          {ObjectId Function(Map<String, dynamic> item)? id,
          String? name,
          String? settingsId,
          bool settingsIdAsNull = false,
          AccountQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => AccountSnap.fromJson(data,
              id: id?.call(data),
              name: name,
              settingsId: settingsId,
              settingsIdAsNull: settingsIdAsNull,
              withQuery: withQuery))
          .toList();
  AccountSnap._(ObjectId id, this.name, this.settingsId, [ObjectId? _updateId])
      : super(id, _updateId);

  final String name;
  final String? settingsId;

  @override
  int compareTo(AccountSnap other) =>
      compare(name, other.name) ?? compare(settingsId, other.settingsId) ?? 0;

  @override
  AccountSnap copyWith(
          {ObjectId? id,
          String? name,
          String? settingsId,
          bool settingsIdAsNull = false}) =>
      AccountSnap._(id ?? this.id, name ?? this.name,
          settingsIdAsNull ? null : (settingsId ?? this.settingsId));

  @override
  Map<String, dynamic> toJson() => {
        'id': id.hexString,
        'name': name,
        if (settingsId != null) 'settings': settingsId
      };
}

class AccountRef extends ModelRef<AccountSnap, ObjectId, AccountRef>
    implements Account, AccountKey {
  final _ModelFeature _feature;
  AccountRef._(this._feature, ObjectId id) : super(_feature, id);

  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);
  SettingsRef? get settings =>
      (_feature.get(this).settingsId != null) ? SettingsRef._(_feature) : null;
  String? get settingsId => _feature.get(this).settingsId;
  set settings(SettingsRef? value) =>
      (value != null) ? update(settings: value) : update(settingsAsNull: true);
  void update(
      {String? name, SettingsKey? settings, bool settingsAsNull = false}) {
    final s = AccountSnap._(id, (name ?? this.name),
        settingsAsNull ? null : (settings ?? this.settings)?.id);
    _feature.put(s);
  }
}

class AccountQuery {
  final ObjectId? id;
  final String? name;
  final String? settingsId;
  final bool? settingsIdIsNull;

  const AccountQuery(
      {this.id, this.name, this.settingsIdIsNull, this.settingsId});
  AccountQuery._(this.id, this.name, this.settingsIdIsNull, this.settingsId);

  bool get isAll =>
      (id == null) &&
      (name == null) &&
      (settingsIdIsNull == null) &&
      (settingsId == null);
  bool get isNotAll => !isAll;

  bool test(AccountSnap m) =>
      (id == null || id == m.id) &&
      (name == null || name == m.name) &&
      ((settingsIdIsNull == true)
          ? (m.settingsId == null)
          : (settingsIdIsNull == false)
              ? (settingsId == null)
                  ? (m.settingsId != null)
                  : (settingsId == m.settingsId)
              : ((settingsId == null) || (settingsId == m.settingsId)));

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is AccountQuery &&
          id == other.id &&
          name == other.name &&
          settingsId == other.settingsId);

  @override
  int get hashCode => Object.hash(Account, id, name, settingsId);
}

abstract class Image {
  int get id;
  String? get name;
  int get height;
  int get width;
}

class ImageKey extends StateKey<ImageSnap> {
  @override
  final int id;
  const ImageKey(this.id);
  ImageKey.parse(String s) : id = ImageKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? ImageKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class ImageSnap extends Model<ImageSnap, int> implements Image, ImageKey {
  ImageSnap(
      {required int id, String? name, required int height, required int width})
      : this._(id, name, height, width);
  ImageSnap.fromJson(Map<String, dynamic> data,
      {int? id,
      String? name,
      bool nameAsNull = false,
      int? height,
      int? width,
      ImageQuery? withQuery})
      : this._(
            id ?? (data['id'] as int),
            name ??
                withQuery?.name ??
                ((data['name'] != null) ? data['name'] : null),
            height ?? withQuery?.height ?? data['height'],
            width ?? withQuery?.width ?? data['width']);
  static List<ImageSnap> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          String? name,
          bool nameAsNull = false,
          int? height,
          int? width,
          ImageQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => ImageSnap.fromJson(data,
              id: id?.call(data),
              name: name,
              nameAsNull: nameAsNull,
              height: height,
              width: width,
              withQuery: withQuery))
          .toList();
  ImageSnap._(int id, this.name, this.height, this.width, [ObjectId? _updateId])
      : super(id, _updateId);

  final String? name;
  final int height;
  final int width;

  @override
  int compareTo(ImageSnap other) =>
      compare(name, other.name) ??
      compare(height, other.height) ??
      compare(width, other.width) ??
      0;

  @override
  ImageSnap copyWith(
          {int? id,
          String? name,
          bool nameAsNull = false,
          int? height,
          int? width}) =>
      ImageSnap._(id ?? this.id, nameAsNull ? null : (name ?? this.name),
          height ?? this.height, width ?? this.width);

  @override
  Map<String, dynamic> toJson() => {
        'id': id,
        if (name != null) 'name': name,
        'height': height,
        'width': width
      };
}

class ImageRef extends ModelRef<ImageSnap, int, ImageRef>
    implements Image, ImageKey {
  final _ModelFeature _feature;
  ImageRef._(this._feature, int id) : super(_feature, id);

  String? get name => _feature.get(this).name;
  set name(String? value) =>
      (value != null) ? update(name: value) : update(nameAsNull: true);
  int get height => _feature.get(this).height;
  set height(int value) => update(height: value);
  int get width => _feature.get(this).width;
  set width(int value) => update(width: value);

  void expire() =>
      _feature<_RestFeature>().imageRestAdapter.expire(ImageQuery(id: id));

  Future<ImageRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .imageRestAdapter
      .fetch(ImageQuery(id: id), force)
      .then((_) => this);

  void update(
      {String? name, bool nameAsNull = false, int? height, int? width}) {
    final s = ImageSnap._(id, nameAsNull ? null : (name ?? this.name),
        (height ?? this.height), (width ?? this.width));
    _feature.put(s);
  }
}

class ImageQuery {
  final int? id;
  final String? name;
  final bool? nameIsNull;
  final int? height;
  final int? width;

  const ImageQuery(
      {this.id, this.nameIsNull, this.name, this.height, this.width});
  ImageQuery._(this.id, this.nameIsNull, this.name, this.height, this.width);

  bool get isAll =>
      (id == null) &&
      (nameIsNull == null) &&
      (name == null) &&
      (height == null) &&
      (width == null);
  bool get isNotAll => !isAll;

  bool test(ImageSnap m) =>
      (id == null || id == m.id) &&
      ((nameIsNull == true)
          ? (m.name == null)
          : (nameIsNull == false)
              ? (name == null)
                  ? (m.name != null)
                  : (name == m.name)
              : ((name == null) || (name == m.name))) &&
      (height == null || height == m.height) &&
      (width == null || width == m.width);

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is ImageQuery &&
          id == other.id &&
          name == other.name &&
          height == other.height &&
          width == other.width);

  @override
  int get hashCode => Object.hash(Image, id, name, height, width);
}

abstract class User {
  int get id;
  String get name;
  String? get desc;
  Address get home;
  Address? get work;
}

class UserKey extends StateKey<UserSnap> {
  @override
  final int id;
  const UserKey(this.id);
  UserKey.parse(String s) : id = UserKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? UserKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class UserSnap extends Model<UserSnap, int> implements User, UserKey {
  UserSnap(
      {required int id,
      required String name,
      String? desc,
      required Address home,
      Address? work})
      : this._(id, name, desc, home, work);
  UserSnap.fromJson(Map<String, dynamic> data,
      {int? id,
      String? name,
      String? desc,
      bool descAsNull = false,
      Address? home,
      Address? work,
      bool workAsNull = false,
      UserQuery? withQuery})
      : this._(
            id ?? (data['id'] as int),
            name ?? withQuery?.name ?? data['name'],
            desc ??
                withQuery?.desc ??
                ((data['desc'] != null) ? data['desc'] : null),
            home ?? withQuery?.home ?? Address.fromJson(data['home']),
            work ??
                withQuery?.work ??
                ((data['work'] != null)
                    ? Address.fromJson(data['work'])
                    : null));
  static List<UserSnap> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          String? name,
          String? desc,
          bool descAsNull = false,
          Address? home,
          Address? work,
          bool workAsNull = false,
          UserQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => UserSnap.fromJson(data,
              id: id?.call(data),
              name: name,
              desc: desc,
              descAsNull: descAsNull,
              home: home,
              work: work,
              workAsNull: workAsNull,
              withQuery: withQuery))
          .toList();
  UserSnap._(int id, this.name, this.desc, this.home, this.work,
      [ObjectId? _updateId])
      : super(id, _updateId);

  final String name;
  final String? desc;
  final Address home;
  final Address? work;

  @override
  int compareTo(UserSnap other) =>
      compare(name, other.name) ??
      compare(desc, other.desc) ??
      compare(home, other.home) ??
      compare(work, other.work) ??
      0;

  @override
  UserSnap copyWith(
          {int? id,
          String? name,
          String? desc,
          bool descAsNull = false,
          Address? home,
          Address? work,
          bool workAsNull = false}) =>
      UserSnap._(
          id ?? this.id,
          name ?? this.name,
          descAsNull ? null : (desc ?? this.desc),
          home ?? this.home,
          workAsNull ? null : (work ?? this.work));

  @override
  Map<String, dynamic> toJson() => {
        'id': id,
        'name': name,
        if (desc != null) 'desc': desc,
        'home': home.toJson(),
        if (work != null) 'work': work?.toJson()
      };
}

class UserRef extends ModelRef<UserSnap, int, UserRef>
    implements User, UserKey {
  final _ModelFeature _feature;
  UserRef._(this._feature, int id) : super(_feature, id);

  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);
  String? get desc => _feature.get(this).desc;
  set desc(String? value) =>
      (value != null) ? update(desc: value) : update(descAsNull: true);
  Address get home => _feature.get(this).home;
  set home(Address value) => update(home: value);
  Address? get work => _feature.get(this).work;
  set work(Address? value) =>
      (value != null) ? update(work: value) : update(workAsNull: true);
  UserPosts get posts => UserPosts._(_feature, this);

  void expire() =>
      _feature<_RestFeature>().userRestAdapter.expire(UserQuery(id: id));

  Future<UserRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .userRestAdapter
      .fetch(UserQuery(id: id), force)
      .then((_) => this);

  void update(
      {String? name,
      String? desc,
      bool descAsNull = false,
      Address? home,
      Address? work,
      bool workAsNull = false}) {
    final s = UserSnap._(
        id,
        (name ?? this.name),
        descAsNull ? null : (desc ?? this.desc),
        (home ?? this.home),
        workAsNull ? null : (work ?? this.work));
    _feature.put(s);
  }
}

class UserQuery {
  final int? id;
  final String? name;
  final String? desc;
  final bool? descIsNull;
  final Address? home;
  final Address? work;
  final bool? workIsNull;

  const UserQuery(
      {this.id,
      this.name,
      this.descIsNull,
      this.desc,
      this.home,
      this.workIsNull,
      this.work});
  UserQuery._(this.id, this.name, this.descIsNull, this.desc, this.home,
      this.workIsNull, this.work);

  bool get isAll =>
      (id == null) &&
      (name == null) &&
      (descIsNull == null) &&
      (desc == null) &&
      (home == null) &&
      (workIsNull == null) &&
      (work == null);
  bool get isNotAll => !isAll;

  bool test(UserSnap m) =>
      (id == null || id == m.id) &&
      (name == null || name == m.name) &&
      ((descIsNull == true)
          ? (m.desc == null)
          : (descIsNull == false)
              ? (desc == null)
                  ? (m.desc != null)
                  : (desc == m.desc)
              : ((desc == null) || (desc == m.desc))) &&
      (home == null || home == m.home) &&
      ((workIsNull == true)
          ? (m.work == null)
          : (workIsNull == false)
              ? (work == null)
                  ? (m.work != null)
                  : (work == m.work)
              : ((work == null) || (work == m.work)));

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is UserQuery &&
          id == other.id &&
          name == other.name &&
          desc == other.desc &&
          home == other.home &&
          work == other.work);

  @override
  int get hashCode => Object.hash(User, id, name, desc, home, work);
}

abstract class Post {
  int get id;
  String get title;
  int get userId;
}

class PostKey extends StateKey<PostSnap> {
  @override
  final int id;
  const PostKey(this.id);
  PostKey.parse(String s) : id = PostKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? PostKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class PostSnap extends Model<PostSnap, int> implements Post, PostKey {
  PostSnap({required int id, required String title, required int userId})
      : this._(id, title, userId);
  PostSnap.fromJson(Map<String, dynamic> data,
      {int? id, String? title, int? userId, PostQuery? withQuery})
      : this._(
            id ?? (data['id'] as int),
            title ?? withQuery?.title ?? data['title'],
            userId ?? withQuery?.userId ?? data['user']);
  static List<PostSnap> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          String? title,
          int? userId,
          PostQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => PostSnap.fromJson(data,
              id: id?.call(data),
              title: title,
              userId: userId,
              withQuery: withQuery))
          .toList();
  PostSnap._(int id, this.title, this.userId, [ObjectId? _updateId])
      : super(id, _updateId);

  final String title;
  final int userId;

  @override
  int compareTo(PostSnap other) =>
      compare(title, other.title) ?? compare(userId, other.userId) ?? 0;

  @override
  PostSnap copyWith({int? id, String? title, int? userId}) =>
      PostSnap._(id ?? this.id, title ?? this.title, userId ?? this.userId);

  @override
  Map<String, dynamic> toJson() => {'id': id, 'title': title, 'user': userId};
}

class PostRef extends ModelRef<PostSnap, int, PostRef>
    implements Post, PostKey {
  final _ModelFeature _feature;
  PostRef._(this._feature, int id) : super(_feature, id);

  String get title => _feature.get(this).title;
  set title(String value) => update(title: value);
  UserRef get user => UserRef._(_feature, _feature.get(this).userId);
  int get userId => _feature.get(this).userId;
  set user(UserRef value) => update(user: value);
  PostComments get comments => PostComments._(_feature, this);

  void expire() =>
      _feature<_RestFeature>().postRestAdapter.expire(PostQuery(id: id));

  Future<PostRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .postRestAdapter
      .fetch(PostQuery(id: id), force)
      .then((_) => this);

  void update({String? title, UserKey? user}) {
    final s = PostSnap._(id, (title ?? this.title), (user ?? this.user).id);
    _feature.put(s);
  }
}

class PostQuery {
  final int? id;
  final String? title;
  final int? userId;

  const PostQuery({this.id, this.title, this.userId});
  PostQuery._(this.id, this.title, this.userId);

  bool get isAll => (id == null) && (title == null) && (userId == null);
  bool get isNotAll => !isAll;

  bool test(PostSnap m) =>
      (id == null || id == m.id) &&
      (title == null || title == m.title) &&
      (userId == null || userId == m.userId);

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is PostQuery &&
          id == other.id &&
          title == other.title &&
          userId == other.userId);

  @override
  int get hashCode => Object.hash(Post, id, title, userId);
}

abstract class Comment {
  int get id;
  int get postId;
  String get message;
  int get ownerId;
  int? get auditorId;
}

class CommentKey extends StateKey<CommentSnap> {
  @override
  final int id;
  const CommentKey(this.id);
  CommentKey.parse(String s) : id = CommentKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? CommentKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class CommentSnap extends Model<CommentSnap, int>
    implements Comment, CommentKey {
  CommentSnap(
      {required int id,
      required int postId,
      required String message,
      required int ownerId,
      int? auditorId})
      : this._(id, postId, message, ownerId, auditorId);
  CommentSnap.fromJson(Map<String, dynamic> data,
      {int? id,
      int? postId,
      String? message,
      int? ownerId,
      int? auditorId,
      bool auditorIdAsNull = false,
      CommentQuery? withQuery})
      : this._(
            id ?? (data['id'] as int),
            postId ?? withQuery?.postId ?? data['post'],
            message ?? withQuery?.message ?? data['message'],
            ownerId ?? withQuery?.ownerId ?? data['owner'],
            auditorId ??
                withQuery?.auditorId ??
                ((data['auditor'] != null) ? data['auditor'] : null));
  static List<CommentSnap> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          int? postId,
          String? message,
          int? ownerId,
          int? auditorId,
          bool auditorIdAsNull = false,
          CommentQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => CommentSnap.fromJson(data,
              id: id?.call(data),
              postId: postId,
              message: message,
              ownerId: ownerId,
              auditorId: auditorId,
              auditorIdAsNull: auditorIdAsNull,
              withQuery: withQuery))
          .toList();
  CommentSnap._(int id, this.postId, this.message, this.ownerId, this.auditorId,
      [ObjectId? _updateId])
      : super(id, _updateId);

  final int postId;
  final String message;
  final int ownerId;
  final int? auditorId;

  @override
  int compareTo(CommentSnap other) =>
      compare(postId, other.postId) ??
      compare(message, other.message) ??
      compare(ownerId, other.ownerId) ??
      compare(auditorId, other.auditorId) ??
      0;

  @override
  CommentSnap copyWith(
          {int? id,
          int? postId,
          String? message,
          int? ownerId,
          int? auditorId,
          bool auditorIdAsNull = false}) =>
      CommentSnap._(
          id ?? this.id,
          postId ?? this.postId,
          message ?? this.message,
          ownerId ?? this.ownerId,
          auditorIdAsNull ? null : (auditorId ?? this.auditorId));

  @override
  Map<String, dynamic> toJson() => {
        'id': id,
        'post': postId,
        'message': message,
        'owner': ownerId,
        if (auditorId != null) 'auditor': auditorId
      };
}

class CommentRef extends ModelRef<CommentSnap, int, CommentRef>
    implements Comment, CommentKey {
  final _ModelFeature _feature;
  CommentRef._(this._feature, int id) : super(_feature, id);

  PostRef get post => PostRef._(_feature, _feature.get(this).postId);
  int get postId => _feature.get(this).postId;
  set post(PostRef value) => update(post: value);
  String get message => _feature.get(this).message;
  set message(String value) => update(message: value);
  UserRef get owner => UserRef._(_feature, _feature.get(this).ownerId);
  int get ownerId => _feature.get(this).ownerId;
  set owner(UserRef value) => update(owner: value);
  UserRef? get auditor {
    final id = _feature.get(this).auditorId;
    return (id != null) ? UserRef._(_feature, id) : null;
  }

  int? get auditorId => _feature.get(this).auditorId;
  set auditor(UserRef? value) =>
      (value != null) ? update(auditor: value) : update(auditorAsNull: true);

  void expire() =>
      _feature<_RestFeature>().commentRestAdapter.expire(CommentQuery(id: id));

  Future<CommentRef> fetch({bool force = false}) => _feature<_RestFeature>()
      .commentRestAdapter
      .fetch(CommentQuery(id: id), force)
      .then((_) => this);

  void update(
      {PostKey? post,
      String? message,
      UserKey? owner,
      UserKey? auditor,
      bool auditorAsNull = false}) {
    final s = CommentSnap._(
        id,
        (post ?? this.post).id,
        (message ?? this.message),
        (owner ?? this.owner).id,
        auditorAsNull ? null : (auditor ?? this.auditor)?.id);
    _feature.put(s);
  }
}

class CommentQuery {
  final int? id;
  final int? postId;
  final String? message;
  final int? ownerId;
  final int? auditorId;
  final bool? auditorIdIsNull;

  const CommentQuery(
      {this.id,
      this.postId,
      this.message,
      this.ownerId,
      this.auditorIdIsNull,
      this.auditorId});
  CommentQuery._(this.id, this.postId, this.message, this.ownerId,
      this.auditorIdIsNull, this.auditorId);

  bool get isAll =>
      (id == null) &&
      (postId == null) &&
      (message == null) &&
      (ownerId == null) &&
      (auditorIdIsNull == null) &&
      (auditorId == null);
  bool get isNotAll => !isAll;

  bool test(CommentSnap m) =>
      (id == null || id == m.id) &&
      (postId == null || postId == m.postId) &&
      (message == null || message == m.message) &&
      (ownerId == null || ownerId == m.ownerId) &&
      ((auditorIdIsNull == true)
          ? (m.auditorId == null)
          : (auditorIdIsNull == false)
              ? (auditorId == null)
                  ? (m.auditorId != null)
                  : (auditorId == m.auditorId)
              : ((auditorId == null) || (auditorId == m.auditorId)));

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is CommentQuery &&
          id == other.id &&
          postId == other.postId &&
          message == other.message &&
          ownerId == other.ownerId &&
          auditorId == other.auditorId);

  @override
  int get hashCode =>
      Object.hash(Comment, id, postId, message, ownerId, auditorId);
}

class Accounts
    extends ModelCollection<Accounts, AccountSnap, ObjectId, AccountRef> {
  final _ModelFeature _feature;
  Accounts._(this._feature)
      : super(_feature, () => Accounts._(_feature),
            (s) => AccountRef._(_feature, s.id));
  AccountRef? operator [](ObjectId id) => getOrNull(id);
  AccountRef get(ObjectId id) =>
      getOrNull(id) ?? (throw Exception('Account($id) does not exist'));
  AccountRef? getOrNull(ObjectId id) {
    final ref = AccountRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Iterable<AccountRef> find(
          {ObjectId? id,
          String? name,
          bool? settingsIdIsNull,
          String? settingsId}) =>
      _feature
          .getAll(AccountQuery._(id, name, settingsIdIsNull, settingsId).test)
          .map((s) => AccountRef._(_feature, s.id));

  AccountRef put({ObjectId? id, required String name, String? settingsId}) =>
      load(AccountSnap._(id ?? ObjectId(), name, settingsId));
}

class Images extends ModelCollection<Images, ImageSnap, int, ImageRef> {
  final _ModelFeature _feature;
  Images._(this._feature)
      : super(_feature, () => Images._(_feature),
            (s) => ImageRef._(_feature, s.id));
  ImageRef? operator [](int id) => getOrNull(id);
  ImageRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Image($id) does not exist'));
  ImageRef? getOrNull(int id) {
    final ref = ImageRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  void expire(
          {int? id, bool? nameIsNull, String? name, int? height, int? width}) =>
      _feature<_RestFeature>()
          .imageRestAdapter
          .expire(ImageQuery._(id, nameIsNull, name, height, width));

  Future<void> fetch(
          {int? id,
          bool? nameIsNull,
          String? name,
          int? height,
          int? width,
          bool force = false}) =>
      _feature<_RestFeature>()
          .imageRestAdapter
          .fetch(ImageQuery._(id, nameIsNull, name, height, width), force);

  Iterable<ImageRef> find(
          {int? id, bool? nameIsNull, String? name, int? height, int? width}) =>
      _feature
          .getAll(ImageQuery._(id, nameIsNull, name, height, width).test)
          .map((s) => ImageRef._(_feature, s.id));

  ImageRef put(
          {required int id,
          String? name,
          required int height,
          required int width}) =>
      load(ImageSnap._(id, name, height, width));
}

class Users extends ModelCollection<Users, UserSnap, int, UserRef> {
  final _ModelFeature _feature;
  Users._(this._feature)
      : super(_feature, () => Users._(_feature),
            (s) => UserRef._(_feature, s.id));
  UserRef? operator [](int id) => getOrNull(id);
  UserRef get(int id) =>
      getOrNull(id) ?? (throw Exception('User($id) does not exist'));
  UserRef? getOrNull(int id) {
    final ref = UserRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  void expire(
          {int? id,
          String? name,
          bool? descIsNull,
          String? desc,
          Address? home,
          bool? workIsNull,
          Address? work}) =>
      _feature<_RestFeature>().userRestAdapter.expire(
          UserQuery._(id, name, descIsNull, desc, home, workIsNull, work));

  Future<void> fetch(
          {int? id,
          String? name,
          bool? descIsNull,
          String? desc,
          Address? home,
          bool? workIsNull,
          Address? work,
          bool force = false}) =>
      _feature<_RestFeature>().userRestAdapter.fetch(
          UserQuery._(id, name, descIsNull, desc, home, workIsNull, work),
          force);

  Iterable<UserRef> find(
          {int? id,
          String? name,
          bool? descIsNull,
          String? desc,
          Address? home,
          bool? workIsNull,
          Address? work}) =>
      _feature
          .getAll(
              UserQuery._(id, name, descIsNull, desc, home, workIsNull, work)
                  .test)
          .map((s) => UserRef._(_feature, s.id));

  UserRef put(
          {required int id,
          required String name,
          String? desc,
          required Address home,
          Address? work}) =>
      load(UserSnap._(id, name, desc, home, work));
}

class UserPosts extends ModelCollection<UserPosts, PostSnap, int, PostRef> {
  final _ModelFeature _feature;
  final UserRef _user;
  UserPosts._(this._feature, this._user)
      : super(_feature, () => UserPosts._(_feature, _user),
            (s) => PostRef._(_feature, s.id), (s) => s.userId == _user.id);
  PostRef? operator [](int id) => getOrNull(id);
  PostRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Post($id) does not exist'));
  PostRef? getOrNull(int id) {
    final ref = PostRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Future<void> fetch({int? id, String? title, bool force = false}) =>
      _feature<_RestFeature>()
          .postRestAdapter
          .fetch(PostQuery._(id, title, _user.id), force);

  Iterable<PostRef> find({int? id, String? title}) => _feature
      .getAll(PostQuery._(id, title, _user.id).test)
      .map((s) => PostRef._(_feature, s.id));

  PostRef put({required int id, required String title}) =>
      load(PostSnap._(id, title, _user.id));
}

class Posts extends ModelCollection<Posts, PostSnap, int, PostRef> {
  final _ModelFeature _feature;
  Posts._(this._feature)
      : super(_feature, () => Posts._(_feature),
            (s) => PostRef._(_feature, s.id));
  PostRef? operator [](int id) => getOrNull(id);
  PostRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Post($id) does not exist'));
  PostRef? getOrNull(int id) {
    final ref = PostRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  void expire({int? id, String? title, int? userId}) => _feature<_RestFeature>()
      .postRestAdapter
      .expire(PostQuery._(id, title, userId));

  Future<void> fetch(
          {int? id, String? title, int? userId, bool force = false}) =>
      _feature<_RestFeature>()
          .postRestAdapter
          .fetch(PostQuery._(id, title, userId), force);

  Iterable<PostRef> find({int? id, String? title, int? userId}) => _feature
      .getAll(PostQuery._(id, title, userId).test)
      .map((s) => PostRef._(_feature, s.id));

  PostRef put({required int id, required String title, required int userId}) =>
      load(PostSnap._(id, title, userId));
}

class PostComments
    extends ModelCollection<PostComments, CommentSnap, int, CommentRef> {
  final _ModelFeature _feature;
  final PostRef _post;
  PostComments._(this._feature, this._post)
      : super(_feature, () => PostComments._(_feature, _post),
            (s) => CommentRef._(_feature, s.id), (s) => s.postId == _post.id);
  CommentRef? operator [](int id) => getOrNull(id);
  CommentRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Comment($id) does not exist'));
  CommentRef? getOrNull(int id) {
    final ref = CommentRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Future<void> fetch(
          {int? id,
          String? message,
          int? ownerId,
          bool? auditorIdIsNull,
          int? auditorId,
          bool force = false}) =>
      _feature<_RestFeature>().commentRestAdapter.fetch(
          CommentQuery._(
              id, _post.id, message, ownerId, auditorIdIsNull, auditorId),
          force);

  Iterable<CommentRef> find(
          {int? id,
          String? message,
          int? ownerId,
          bool? auditorIdIsNull,
          int? auditorId}) =>
      _feature
          .getAll(CommentQuery._(
                  id, _post.id, message, ownerId, auditorIdIsNull, auditorId)
              .test)
          .map((s) => CommentRef._(_feature, s.id));

  CommentRef put(
          {required int id,
          required String message,
          required int ownerId,
          int? auditorId}) =>
      load(CommentSnap._(id, _post.id, message, ownerId, auditorId));
}

class Comments extends ModelCollection<Comments, CommentSnap, int, CommentRef> {
  final _ModelFeature _feature;
  Comments._(this._feature)
      : super(_feature, () => Comments._(_feature),
            (s) => CommentRef._(_feature, s.id));
  CommentRef? operator [](int id) => getOrNull(id);
  CommentRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Comment($id) does not exist'));
  CommentRef? getOrNull(int id) {
    final ref = CommentRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  void expire(
          {int? id,
          int? postId,
          String? message,
          int? ownerId,
          bool? auditorIdIsNull,
          int? auditorId}) =>
      _feature<_RestFeature>().commentRestAdapter.expire(CommentQuery._(
          id, postId, message, ownerId, auditorIdIsNull, auditorId));

  Future<void> fetch(
          {int? id,
          int? postId,
          String? message,
          int? ownerId,
          bool? auditorIdIsNull,
          int? auditorId,
          bool force = false}) =>
      _feature<_RestFeature>().commentRestAdapter.fetch(
          CommentQuery._(
              id, postId, message, ownerId, auditorIdIsNull, auditorId),
          force);

  Iterable<CommentRef> find(
          {int? id,
          int? postId,
          String? message,
          int? ownerId,
          bool? auditorIdIsNull,
          int? auditorId}) =>
      _feature
          .getAll(CommentQuery._(
                  id, postId, message, ownerId, auditorIdIsNull, auditorId)
              .test)
          .map((s) => CommentRef._(_feature, s.id));

  CommentRef put(
          {required int id,
          required int postId,
          required String message,
          required int ownerId,
          int? auditorId}) =>
      load(CommentSnap._(id, postId, message, ownerId, auditorId));
}

class _ModelFeature extends WatchableStateFeature {
  late final settings = () {
    if (!contains(const SettingsKey())) {
      put(SettingsSnap._('initial settings name', null, null));
    }
    return SettingsRef._(this);
  }();
  late final config = ConfigRef._(this);
  late final accounts = Accounts._(this);
  late final images = Images._(this);
  late final users = Users._(this);
  late final posts = Posts._(this);
  late final comments = Comments._(this);
  void load(data) {
    if (data is SettingsSnap) {
      put(data);
    } else if (data is ConfigSnap) {
      put(data);
    } else if (data is AccountSnap) {
      put(data);
    } else if (data is ImageSnap) {
      put(data);
    } else if (data is UserSnap) {
      put(data);
    } else if (data is PostSnap) {
      put(data);
    } else if (data is CommentSnap) {
      put(data);
    } else if (data is Iterable) {
      data.forEach((e) => load(e));
    } else if (data is Map<String, dynamic>) {
      for (final key in data.keys) {
        if (key == 'settings' && data[key] is Map<String, dynamic>) {
          load(SettingsSnap.fromJson(data['settings']));
        } else if (key == 'config' && data[key] is Map<String, dynamic>) {
          load(ConfigSnap.fromJson(data['config']));
        } else if (key == 'account' && data[key] is Map<String, dynamic>) {
          load(AccountSnap.fromJson(data['account']));
        } else if (key == 'image' && data[key] is Map<String, dynamic>) {
          load(ImageSnap.fromJson(data['image']));
        } else if (key == 'user' && data[key] is Map<String, dynamic>) {
          load(UserSnap.fromJson(data['user']));
        } else if (key == 'post' && data[key] is Map<String, dynamic>) {
          load(PostSnap.fromJson(data['post']));
        } else if (key == 'comment' && data[key] is Map<String, dynamic>) {
          load(CommentSnap.fromJson(data['comment']));
        } else if (key == 'accounts' && data[key] is Iterable) {
          load(AccountSnap.fromJsonList(data['accounts']));
        } else if (key == 'images' && data[key] is Iterable) {
          load(ImageSnap.fromJsonList(data['images']));
        } else if (key == 'users' && data[key] is Iterable) {
          load(UserSnap.fromJsonList(data['users']));
        } else if (key == 'posts' && data[key] is Iterable) {
          load(PostSnap.fromJsonList(data['posts']));
        } else if (key == 'comments' && data[key] is Iterable) {
          load(CommentSnap.fromJsonList(data['comments']));
        } else {
          throw Exception(
              'invalid map entry during load: ($key, ${data[key]})');
        }
      }
    } else {
      throw Exception('attempted to load unknown type: $data');
    }
  }

  void unload(data) {
    if (data is SettingsKey) {
      remove(data);
    } else if (data is ConfigKey) {
      remove(data);
    } else if (data is AccountKey) {
      remove(data);
    } else if (data is ImageKey) {
      remove(data);
    } else if (data is UserKey) {
      remove(data);
    } else if (data is PostKey) {
      remove(data);
    } else if (data is CommentKey) {
      remove(data);
    } else if (data is Iterable) {
      data.forEach((e) => unload(e));
    } else if (data is Map<String, dynamic>) {
      for (final key in data.keys) {
        if (key == 'settings' && data[key] is String) {
          unload(SettingsKey);
        } else if (key == 'config' && data[key] is String) {
          unload(ConfigKey);
        } else if (key == 'account' && data[key] is String) {
          unload(AccountKey.parse(data[key]));
        } else if (key == 'image' && data[key] is int) {
          unload(ImageKey(data[key] as int));
        } else if (key == 'user' && data[key] is int) {
          unload(UserKey(data[key] as int));
        } else if (key == 'post' && data[key] is int) {
          unload(PostKey(data[key] as int));
        } else if (key == 'comment' && data[key] is int) {
          unload(CommentKey(data[key] as int));
        } else if (key == 'accounts' && data[key] is Iterable<String>) {
          (data[key] as Iterable)
              .forEach((e) => unload(AccountKey.parse(data[key])));
        } else if (key == 'images' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(ImageKey(data[key] as int)));
        } else if (key == 'users' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(UserKey(data[key] as int)));
        } else if (key == 'posts' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(PostKey(data[key] as int)));
        } else if (key == 'comments' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(CommentKey(data[key] as int)));
        } else {
          throw Exception(
              'invalid map entry during unload: ($key, ${data[key]})');
        }
      }
    } else {
      throw Exception('attempted to unload unknown type: $data');
    }
  }

  @override
  void onEvents(StateEvents events) {
    for (final v in events.values) {
      if (v is SettingsSnap) {
        if (v.accountId != null) {
          _onRipple(AccountKey(v.accountId!));
        }
        if (v.userId != null) {
          _onRipple(UserKey(v.userId!));
        }
      } else if (v is AccountSnap) {
        if (v.settingsId != null) {
          _onRipple(SettingsKey());
        }
      } else if (v is PostSnap) {
        _onRipple(UserKey(v.userId));
      } else if (v is CommentSnap) {
        _onRipple(PostKey(v.postId));
        _onRipple(UserKey(v.ownerId));
        if (v.auditorId != null) {
          _onRipple(UserKey(v.auditorId!));
        }
      }
    }
    super.onEvents(events);
  }

  void _onRipple(StateKey k) {
    if (addRipple(k)) {
      if (k is SettingsKey) {
        final settings = getOrNull(k);
        final settings_account = settings?.accountId;
        if (settings_account != null) {
          _onRipple(AccountKey(settings_account));
        }
        final settings_user = settings?.userId;
        if (settings_user != null) {
          _onRipple(UserKey(settings_user));
        }
      } else if (k is AccountKey) {
        final account = getOrNull(k);
        final account_settings = account?.settingsId;
        if (account_settings != null) {
          _onRipple(SettingsKey());
        }
      } else if (k is PostKey) {
        final post = getOrNull(k);
        final post_user = post?.userId;
        if (post_user != null) {
          _onRipple(UserKey(post_user));
        }
      } else if (k is CommentKey) {
        final comment = getOrNull(k);
        final comment_post = comment?.postId;
        if (comment_post != null) {
          _onRipple(PostKey(comment_post));
        }
        final comment_owner = comment?.ownerId;
        if (comment_owner != null) {
          _onRipple(UserKey(comment_owner));
        }
        final comment_auditor = comment?.auditorId;
        if (comment_auditor != null) {
          _onRipple(UserKey(comment_auditor));
        }
      }
    }
  }
}
