import 'dart:async';
import 'package:oobium_datastore/oobium_datastore.dart';
import 'package:oobium_logging/oobium_logging.dart';
part 'models_test.z.schema.g.feature_persist.dart';
part 'models_test.z.schema.x.dart';

final _log = Logger(ModelsTestData);

class ModelsTestData extends ModelContext {
  ModelsTestData({String? path, OnUpgrade? onUpgrade})
      : super(
            StateContext(features: [
              _ModelFeature(),
              if (path != null)
                _PersistFeature(
                    path: path,
                    version: 3,
                    compactionStrategy: const DefaultCompactionStrategy(),
                    converter: const _Converter(),
                    onUpgrade: onUpgrade ?? dropOnUpgrade)
            ]),
            const _Converter());

  Future<ModelsTestData> open() => super.open().then((_) => this);

  Images get images => feature<_ModelFeature>().images;

  void load(data) => feature<_ModelFeature>().load(data);
  void unload(data) => feature<_ModelFeature>().unload(data);
}

class _Converter extends Converter {
  const _Converter();

  @override
  StateValue decodePut(DataRecord r) {
    final data = r.toJsonData();
    switch (r.type) {
      case 'Image':
        return ImageSnap._(ImageKey.value(r.modelId), data['name'],
            data['height'], data['width'], ObjectId.fromHexString(r.updateId));
      default:
        throw Exception('unknown type ${r.type} during decodePut');
    }
  }

  @override
  StateKey decodeRemove(DataRecord r) {
    switch (r.type) {
      case 'Image':
        return ImageKey.parse(r.modelId);
      default:
        throw Exception('unknown type ${r.type} during decodeRemove');
    }
  }

  @override
  DataRecord encodePut(StateValue v) {
    if (v is ImageSnap) {
      return DataRecord('Image', v.id.toString(), v.updateId.hexString,
          {'name': v.name, 'height': v.height, 'width': v.width});
    } else {
      throw Exception('unknown type $v during encodePut');
    }
  }

  @override
  DataRecord encodeRemove(StateValue v) {
    if (v is ImageSnap) {
      return DataRecord('Image', v.id.toString(), v.updateId.hexString);
    } else {
      throw Exception('unknown type $v during encodeRemove');
    }
  }
}

abstract class Image {
  int get id;
  String get name;
  int get height;
  int get width;
}

class ImageKey extends StateKey<ImageSnap> {
  @override
  final int id;
  const ImageKey(this.id);
  ImageKey.parse(String s) : id = ImageKey.value(s);
  static int value(String s) => int.parse(s);
  static int? maybe(String? s) {
    try {
      return (s != null) ? ImageKey.value(s) : null;
    } catch (_) {
      return null;
    }
  }
}

class ImageSnap extends Model<ImageSnap, int> implements Image, ImageKey {
  ImageSnap(
      {required int id,
      required String name,
      required int height,
      required int width})
      : this._(id, name, height, width);
  ImageSnap.fromJson(Map<String, dynamic> data,
      {int? id, String? name, int? height, int? width, ImageQuery? withQuery})
      : this._(
            id ?? (data['id'] as int),
            name ?? withQuery?.name ?? data['name'],
            height ?? withQuery?.height ?? data['height'],
            width ?? withQuery?.width ?? data['width']);
  static List<ImageSnap> fromJsonList(list,
          {int Function(Map<String, dynamic> item)? id,
          String? name,
          int? height,
          int? width,
          ImageQuery? withQuery}) =>
      (list as Iterable)
          .map((data) => ImageSnap.fromJson(data,
              id: id?.call(data),
              name: name,
              height: height,
              width: width,
              withQuery: withQuery))
          .toList();
  ImageSnap._(int id, this.name, this.height, this.width, [ObjectId? _updateId])
      : super(id, _updateId);

  final String name;
  final int height;
  final int width;

  @override
  int compareTo(ImageSnap other) =>
      compare(name, other.name) ??
      compare(height, other.height) ??
      compare(width, other.width) ??
      0;

  @override
  ImageSnap copyWith({int? id, String? name, int? height, int? width}) =>
      ImageSnap._(id ?? this.id, name ?? this.name, height ?? this.height,
          width ?? this.width);

  @override
  Map<String, dynamic> toJson() =>
      {'id': id, 'name': name, 'height': height, 'width': width};
}

class ImageRef extends ModelRef<ImageSnap, int, ImageRef>
    implements Image, ImageKey {
  final _ModelFeature _feature;
  ImageRef._(this._feature, int id) : super(_feature, id);

  String get name => _feature.get(this).name;
  set name(String value) => update(name: value);
  int get height => _feature.get(this).height;
  set height(int value) => update(height: value);
  int get width => _feature.get(this).width;
  set width(int value) => update(width: value);

  void update({String? name, int? height, int? width}) {
    final s = ImageSnap._(id, (name ?? this.name), (height ?? this.height),
        (width ?? this.width));
    _feature.put(s);
  }
}

class ImageQuery {
  final int? id;
  final String? name;
  final int? height;
  final int? width;

  const ImageQuery({this.id, this.name, this.height, this.width});
  ImageQuery._(this.id, this.name, this.height, this.width);

  bool get isAll =>
      (id == null) && (name == null) && (height == null) && (width == null);
  bool get isNotAll => !isAll;

  bool test(ImageSnap m) =>
      (id == null || id == m.id) &&
      (name == null || name == m.name) &&
      (height == null || height == m.height) &&
      (width == null || width == m.width);

  @override
  bool operator ==(Object? other) =>
      identical(this, other) ||
      (other is ImageQuery &&
          id == other.id &&
          name == other.name &&
          height == other.height &&
          width == other.width);

  @override
  int get hashCode => Object.hash(Image, id, name, height, width);
}

class Images extends ModelCollection<Images, ImageSnap, int, ImageRef> {
  final _ModelFeature _feature;
  Images._(this._feature)
      : super(_feature, () => Images._(_feature),
            (s) => ImageRef._(_feature, s.id));
  ImageRef? operator [](int id) => getOrNull(id);
  ImageRef get(int id) =>
      getOrNull(id) ?? (throw Exception('Image($id) does not exist'));
  ImageRef? getOrNull(int id) {
    final ref = ImageRef._(_feature, id);
    return _feature.contains(ref) ? ref : null;
  }

  Iterable<ImageRef> find({int? id, String? name, int? height, int? width}) =>
      _feature
          .getAll(ImageQuery._(id, name, height, width).test)
          .map((s) => ImageRef._(_feature, s.id));

  ImageRef put(
          {required int id,
          required String name,
          required int height,
          required int width}) =>
      load(ImageSnap._(id, name, height, width));
}

class _ModelFeature extends WatchableStateFeature {
  late final images = Images._(this);
  void load(data) {
    if (data is ImageSnap) {
      put(data);
    } else if (data is Iterable) {
      data.forEach((e) => load(e));
    } else if (data is Map<String, dynamic>) {
      for (final key in data.keys) {
        if (key == 'image' && data[key] is Map<String, dynamic>) {
          load(ImageSnap.fromJson(data['image']));
        } else if (key == 'images' && data[key] is Iterable) {
          load(ImageSnap.fromJsonList(data['images']));
        } else {
          throw Exception(
              'invalid map entry during load: ($key, ${data[key]})');
        }
      }
    } else {
      throw Exception('attempted to load unknown type: $data');
    }
  }

  void unload(data) {
    if (data is ImageKey) {
      remove(data);
    } else if (data is Iterable) {
      data.forEach((e) => unload(e));
    } else if (data is Map<String, dynamic>) {
      for (final key in data.keys) {
        if (key == 'image' && data[key] is int) {
          unload(ImageKey(data[key] as int));
        } else if (key == 'images' && data[key] is Iterable<int>) {
          (data[key] as Iterable)
              .forEach((e) => unload(ImageKey(data[key] as int)));
        } else {
          throw Exception(
              'invalid map entry during unload: ($key, ${data[key]})');
        }
      }
    } else {
      throw Exception('attempted to unload unknown type: $data');
    }
  }
}
